use std::cmp::min;
use std::collections::{BTreeMap, BTreeSet};
use std::fmt::Debug;
use std::io::Write;
use std::ops::Range;
use std::str::FromStr;
use std::{env, io};

fn parse_map(args: Vec<String>) -> BTreeMap<String, String> {
    let mut res = BTreeMap::new();
    for arg in args {
        if !arg.contains("=") {
            continue;
        }
        let split: Vec<_> = arg.split("=").collect();
        res.insert(split[0].to_owned(), split[1].to_owned());
    }
    res
}

#[allow(dead_code)]
struct Random {
    state: u64,
}

impl Random {
    fn next(&mut self) -> u64 {
        let mut x = self.state;
        x ^= x << 13;
        x ^= x >> 7;
        x ^= x << 17;
        self.state = x;
        x
    }

    #[allow(dead_code)]
    fn next_in_range(&mut self, from: usize, to: usize) -> usize {
        assert!(from < to);
        (from as u64 + self.next() % ((to - from) as u64)) as usize
    }

    #[allow(dead_code)]
    fn next_double(&mut self) -> f64 {
        (self.next() as f64) / (std::usize::MAX as f64)
    }

    #[allow(dead_code)]
    fn new(seed: u64) -> Self {
        assert_ne!(seed, 0);
        Self { state: seed }
    }
}

fn get_arg<T: FromStr>(args: &BTreeMap<String, String>, key: &str) -> T
where
    <T as FromStr>::Err: Debug,
{
    args.get(key).unwrap().parse::<T>().unwrap()
}

fn get_arg_opt<T: FromStr>(args: &BTreeMap<String, String>, key: &str) -> Option<T> {
    args.get(key)?.parse::<T>().ok()
}

struct Test {
    files: Vec<String>,
    k: usize,
}

fn shuffle_files(files: BTreeSet<String>, rnd: &mut Random) -> Vec<String> {
    let mut res: Vec<_> = files.into_iter().collect();
    for i in 1..res.len() {
        let j = rnd.next_in_range(0, i + 1);
        res.swap(i, j);
    }
    res
}

fn gen(args_map: &BTreeMap<String, String>) -> Test {
    let seed: u64 = get_arg(args_map, "seed");
    let mut rnd = Random::new(seed);
    let mut files = BTreeSet::new();
    let k = match get_arg_opt::<usize>(args_map, "k") {
        Some(k) => k,
        None => {
            let max_k: usize = get_arg(args_map, "max_k");
            rnd.next_in_range(1, max_k + 1)
        }
    };
    let max_n: usize = get_arg(args_map, "max_n");
    let max_c: usize = get_arg_opt(args_map, "max_c").unwrap_or(26);
    let test_type: String = get_arg_opt(args_map, "type").unwrap_or("default".to_owned());
    let n = rnd.next_in_range(1, max_n + 1);
    const MAX_SUM_LEN: usize = 300_000;

    let gen_string = |len: usize, rnd: &mut Random| -> Vec<u8> {
        (0..len)
            .map(|_| rnd.next_in_range(0, max_c) as u8 + b'a')
            .collect()
    };

    if test_type == "same_pref" {
        let len = MAX_SUM_LEN / n;
        let default_str = gen_string(len, &mut rnd);
        while files.len() < n {
            let mut my_str = default_str.clone();
            for _ in 0..10 {
                let pos = len - rnd.next_in_range(0, min(10, len)) - 1;
                my_str[pos] = rnd.next_in_range(0, max_c) as u8 + b'a';
            }
            files.insert(String::from_utf8(my_str).unwrap());
        }
    } else {
        assert_eq!(test_type, "default");
        let max_len: usize = get_arg(args_map, "max_len");

        let mut cur_sum_len = 0;
        let mut not_added_for_a_long_time = 0;
        while files.len() < n && cur_sum_len < MAX_SUM_LEN {
            if not_added_for_a_long_time == 100 {
                break;
            }
            let real_max_len = min(max_len, MAX_SUM_LEN - cur_sum_len);
            let use_len = rnd.next_in_range(1, real_max_len + 1);
            let s: Vec<_> = gen_string(use_len, &mut rnd);
            let s = String::from_utf8(s).unwrap();
            if files.contains(&s) {
                not_added_for_a_long_time += 1;
                continue;
            }
            cur_sum_len += s.len();
            files.insert(s);
            not_added_for_a_long_time = 0;
        }
    }
    Test {
        files: shuffle_files(files, &mut rnd),
        k,
    }
}

pub fn main() {
    let stdout = io::stdout();
    let mut out = std::io::BufWriter::new(stdout.lock());

    let args: Vec<String> = env::args().collect();
    let args_map = parse_map(args);
    let test = gen(&args_map);
    // \r is for windows :(
    writeln!(out, "{} {}\r", test.files.len(), test.k).unwrap();
    for file in test.files.iter() {
        writeln!(out, "{}\r", file).unwrap();
    }
    out.flush().unwrap();
}
