#include "testlib.h"
#include <string>
#include <stdexcept>

const int MIN_VALUE = -2147483648;
const int MAX_VALUE = 2147483647;
const int MAX_TOKENS = 1000;

struct command {
    virtual void execute(std::vector<int> &stack, int x, int y) = 0;
    virtual ~command() {}
};

struct push_x_command : command {
    virtual void execute(std::vector<int> &stack, int x, int y) {
        stack.push_back(x);
    }
};

struct push_y_command : command {
    virtual void execute(std::vector<int> &stack, int x, int y) {
        stack.push_back(y);
    }
};

struct push_value_command : command {
    int value;
    push_value_command(int value): value(value) {}
    virtual void execute(std::vector<int> &stack, int x, int y) {
        stack.push_back(value);
    }
};

struct bin_command : command {
    virtual long long apply(long long a, long long b, int x, int y) = 0;
    virtual char name() = 0;
    virtual void execute(std::vector<int> &stack, int x, int y) {
        if (stack.size() < 2) {
            ouf.quitf(_wa, "Command '%c' is applied to stack size %lu", name(), stack.size());
        }
        long long result = stack.back();
        stack.pop_back();
        result = apply(stack.back(), result, x, y);
        stack.pop_back();
        if (result < MIN_VALUE || result > MAX_VALUE) {
            ouf.quitf(_wa, "Command '%c' overflowed at x = %d, y = %d", name(), x, y);
        }
        stack.push_back((int) (result));
    }
};

struct add_command : bin_command {
    virtual long long apply(long long a, long long b, int x, int y) {
        return a + b; // valid input int32s will never overflow int64
    }
    virtual char name() {
        return '+';
    }
};

struct sub_command : bin_command {
    virtual long long apply(long long a, long long b, int x, int y) {
        return a - b; // valid input int32s will never overflow int64
    }
    virtual char name() {
        return '-';
    }
};

struct mul_command : bin_command {
    virtual long long apply(long long a, long long b, int x, int y) {
        return a * b; // valid input int32s will never overflow int64
    }
    virtual char name() {
        return '*';
    }
};

struct pow_command : bin_command {
    virtual long long apply(long long a, long long b, int x, int y) {
        if (b < 0) {
            ouf.quitf(_wa, "Command '^': second argument is negative (%lld)", b);
        }
        long long result = 1; // initially in int32
        long long curr_pow = a; // initially in int32
        while (b > 0) {
            if (curr_pow < MIN_VALUE || curr_pow > MAX_VALUE) {
                ouf.quitf(_wa, "Command '^' overflowed at x = %d, y = %d (path 1)", x, y);
            }
            if (b & 1) {
                // both are in int32 at this moment
                result *= curr_pow;
                if (result < MIN_VALUE || result > MAX_VALUE) {
                    ouf.quitf(_wa, "Command '^' overflowed at x = %d, y = %d (path 2)", x, y);
                }
                // result is in int32
            }
            b >>= 1;
            curr_pow *= curr_pow; // since it was in int32 before, it will not overflow int64
        }
        return result;
    }
    virtual char name() {
        return '^';
    }
};

void read_commands(InStream &ouf, std::vector<command*> &commands) {
    bool first = true;
    while (!ouf.eof()) {
        if (commands.size() == MAX_TOKENS) {
            ouf.quitf(_wa, "Too many tokens");
        }
        std::string cmd = ouf.readToken();
        ouf.skipBlanks(); // well, here we don't really care whether the output is single-space separated...
        if (first && cmd == "No") {
            // the "No solution" path
            if (ouf.eof()) {
                ouf.quitf(_wa, "Just 'No' in the input");
            }
            cmd = ouf.readToken();
            if (cmd == "solution") {
                ouf.quitf(_wa, "Found 'No solution', which cannot happen");
            }
            ouf.quitf(_wa, "Expected 'solution' to follow 'No', '%s' found", cmd.c_str());
        }
        first = false;
        if (cmd == "x") {
            commands.push_back(new push_x_command());
        } else if (cmd == "y") {
            commands.push_back(new push_y_command());
        } else if (cmd == "+") {
            commands.push_back(new add_command());
        } else if (cmd == "-") {
            commands.push_back(new sub_command());
        } else if (cmd == "*") {
            commands.push_back(new mul_command());
        } else if (cmd == "^") {
            commands.push_back(new pow_command());
        } else {
            long long value = stringToLongLong(ouf, cmd.c_str());
            if (value < -9 || value > +9) {
                ouf.quitf(_wa, "Input integer token too large: %lld", value);
            }
            commands.push_back(new push_value_command((int) value));
        }
    }
}

void evaluate_table(std::vector<command*> const &commands, std::vector< std::vector<int> > &table) {
    std::vector<int> stack;
    for (int x = -10; x <= 10; ++x) {
        table.push_back(std::vector<int>());
        for (int y = -10; y <= 10; ++y) {
            stack.clear();
            for (unsigned i = 0; i < commands.size(); ++i) {
                commands[i]->execute(stack, x, y); // each call will either quit or retain stack in a sane form
            }
            if (stack.size() != 1) {
                ouf.quitf(_wa, "Stack size is %lu at the end of evaluation", stack.size());
            }
            table.back().push_back(stack[0]);
        }
    }
}

const int dx[] = {1, 0, -1, 0};
const int dy[] = {0, 1, 0, -1};

bool hasMaxima(std::vector< std::vector<int> > const &table) {
    int countMaxima = 0;
    for (int x = 1; x <= 19; ++x) {
        for (int y = 1; y <= 19; ++y) {
            int v = table[x][y];
            bool isMax = true;
            for (int d = 0; d < 4; ++d) {
                if (table[x + dx[d]][y + dy[d]] >= v) {
                    isMax = false;
                    break;
                }
            }
            if (isMax) {
                ++countMaxima;
            }
        }
    }
    return countMaxima > 1;
}

bool hasMinima(std::vector< std::vector<int> > const &table) {
    int countMinima = 0;
    for (int x = 1; x <= 19; ++x) {
        for (int y = 1; y <= 19; ++y) {
            int v = table[x][y];
            bool isMin = true;
            for (int d = 0; d < 4; ++d) {
                if (table[x + dx[d]][y + dy[d]] <= v) {
                    isMin = false;
                    break;
                }
            }
            if (isMin) {
                ++countMinima;
            }
        }
    }
    return countMinima > 1;
}

bool hasPlateaus(std::vector< std::vector<int> > const &table) {
    for (int x = 0; x <= 20; ++x) {
        for (int y = 0; y <= 20; ++y) {
            int v = table[x][y];
            if ((x < 20 && v == table[x + 1][y]) || (y < 20 && v == table[x][y + 1])) {
                return true;
            }
        }
    }
    return false;
}

char const *yesNo[] = { "No", "Yes" };

int main(int argc, char* argv[]) {
    registerTestlibCmd(argc, argv);

    // Validation of inf is the responsibility of the validator.
    bool inHasMaxima = inf.readString().back() == 's';
    bool inHasMinima = inf.readString().back() == 's';
    bool inHasPlateaus = inf.readString().back() == 's';

    std::vector<command*> commands;
    read_commands(ouf, commands);

    std::vector< std::vector<int> > table;
    evaluate_table(commands, table);

    bool outHasMaxima = hasMaxima(table);
    bool outHasMinima = hasMinima(table);
    bool outHasPlateaus = hasPlateaus(table);

    if (inHasMaxima != outHasMaxima) {
        ouf.quitf(_wa, "Multiple local maxima: expected %s, evaluated to %s", yesNo[inHasMaxima], yesNo[outHasMaxima]);
    }
    if (inHasMinima != outHasMinima) {
        ouf.quitf(_wa, "Multiple local minima: expected %s, evaluated to %s", yesNo[inHasMinima], yesNo[outHasMinima]);
    }
    if (inHasPlateaus != outHasPlateaus) {
        ouf.quitf(_wa, "Plateaus: expected %s, evaluated to %s", yesNo[inHasPlateaus], yesNo[outHasPlateaus]);
    }
    ouf.quitf(_ok, "OK");
}
